PREFIX = """You act like a material scientist answering a question. Answer the following questions as best you can. You have access to the following tools:"""

FORMAT_INSTRUCTIONS = """You must not use tool `generator` when `Question` does not required to generate. 

Use the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Action: the action to take, should be one of [{tool_names}]
Action Input: the input to the action
Observation: the result of the action
... (this Thought/Action/Action Input/Observation can repeat N times)
Thought: I now know the final answer
Final Answer: the final answer to the original input question"""

SUFFIX = """Begin!

Question: What is the surface area and bandgap of ACOGEF?
Thought: I need to find the surface area of ACOGEF.
Action: search_csv
Action Input: "Search name ACOGEF and provide information on its surface area"
Observation: The surface area of material "ACOGEF" is 1138.35 m^2/g. Check to see if this answer can be you final answer, and if so, you should submit your final answer.
Thought: The search_csv tool provided the surface area of ACOGEF, but not the bandgap. I need to find the bandgap.
Action: search_csv
Action Input: "Search name ACOGEF and provide information on its bandgap"
Observation: The search_csv tool did not provide any information on the bandgap of ACOGEF. I need to find another way to obtain this information.
Thought: The search_csv tool provided the surface area of ACOGEF, but not the bandgap. I need to find the bandgap.
Action: predictor
Action Input: "Predict the bandgap of ACOGEF"
Observation: The bandgap of material "ACOGEF" is 3.41139 eV. Check to see if this answer can be you final answer, and if so, you should submit your final answer.
Thought: I now know the final answer
Final Answer: The bandgap and surface area of ACOGEF is 3.41149 eV and 1138.35 m^2/g.

Question: What is the highest bandgap MOF?
Thought: I need to find the MOF with the highest bandgap.
Action: search_csv
Action Input: "find for MOFs with highest bandgap and provide its name and bandgap"
Observation: There are no data in database
Thought: The search_csv tool did not provide any information on the bandgaps of MOFs. I need to find another way to obtain this information.
Action: predictor
Action Input: "predict the bandgaps of MOFs and find the name of MOF with highest bandgaps"
Observation: The highest bandgap MOF is ACOGEF.
Thought: I now know the final answer
Final Answer: The highest bandgap MOF is ACOGEF.

Question: {input}
Thought:{agent_scratchpad}"""


from langchain.chat_models import ChatOpenAI
from langchain.llms import OpenAI
from chatmof.agents.agent import ChatMOF
from langchain.callbacks import StdOutCallbackHandler

import io
import re
from contextlib import redirect_stdout
import pandas as pd
import random
from pathlib import Path

df = pd.read_excel('./database/tables/coremof.xlsx')
names = df['name']


df = pd.read_excel('test/search_csv.xlsx')
questions = list(df[1])

verbose = True
search_internet = False


#llms = []
#for model in ['text-davinci-003', 'text-davinci-002', 'davinci', 'curie', 'babbage', 'ada']:
##    llm = OpenAI(temperature=0, model=model)
#    llms.append([model, llm])

#for model, llm in llms:
llm = ChatOpenAI(temperature=0.1, model_name='gpt-4')
model='gpt-4'

callback_manager = [StdOutCallbackHandler()]

chatmof = ChatMOF.from_llm(
    llm=llm, 
    verbose=verbose, 
)
qna = []

save_path = Path(f'./test/searcher_{model}')
save_path.mkdir(exist_ok=True, parents=True)

for i, question in enumerate(questions):
    question = question.strip()

    with io.StringIO() as buf, redirect_stdout(buf):
        try:
            output = chatmof.run(question, callbacks=callback_manager)
        except Exception as e:
            print (type(e), e)
            output = None
        log = buf.getvalue()

    qna.append([i, question, output])

    with (save_path/f'{i}_.output').open('w') as f:
        f.write(log)

df = pd.DataFrame(qna)
df.to_csv(f'test/search_csv_{model}.csv')


import argparse
from importlib import import_module
from argparse import RawTextHelpFormatter
from chatmof import __version__

commands = [
    ("run", "chatmof.cli.run"),
    ("setup", "chatmof.cli.setup"),
]


def main(prog="chatmof", version=__version__, commands=commands, args=None):
    parser = argparse.ArgumentParser(
        prog=prog,
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s-{}".format(version)
    )
    parser.add_argument("-T", "--traceback", action="store_true")
    subparsers = parser.add_subparsers(title="Sub-command", dest="command")

    subparser = subparsers.add_parser(
        "help", description="Help", help="Help for sub-command."
    )

    subparser.add_argument(
        "helpcommand",
        nargs="?",
        metavar="sub-command",
        help="Provide help for sub-command.",
    )

    functions = {}
    parsers = {}
    for command, module_name in commands:
        cmd = import_module(module_name).CLICommand
        docstring = cmd.__doc__
        if docstring is None:
            # Backwards compatibility with GPAW
            short = cmd.short_description
            long = getattr(cmd, "description", short)
        else:
            parts = docstring.split("\n", 1)
            if len(parts) == 1:
                short = docstring
                long = docstring
            else:
                short, body = parts
                long = short
                # long = short + '\n' + textwrap.dedent(body)
        subparser = subparsers.add_parser(
            command, formatter_class=RawTextHelpFormatter, help=short, description=long
        )
        cmd.add_arguments(subparser)
        functions[command] = cmd.run
        parsers[command] = subparser

    args = parser.parse_args(args)

    if args.command == "help":
        if args.helpcommand is None:
            parser.print_help()
        else:
            parsers[args.helpcommand].print_help()
    elif args.command is None:
        parser.print_usage()
    else:
        f = functions[args.command]
        try:
            if f.__code__.co_argcount == 1:
                f(args)
            else:
                f(args, parsers[args.command])
        except KeyboardInterrupt:
            pass
        except Exception as x:
            if args.traceback:
                raise
            else:
                l1 = "{}: {}\n".format(x.__class__.__name__, x)
                l2 = "To get a full traceback, use: {} -T {} ...".format(
                    prog, args.command
                )
                parser.error(l1 + l2)


from itertools import chain


str_kwargs_names = {
    'accelerator': "Device name for MOFTransformer. accelerator must be one of [cuda, gpu, cpu] (default: cuda)",
    'logger': 'logger for generation task (Default: generate_mof.log)',
}

int_kwargs_names = {
    'max_length_in_predictor': 'max number of MOFs in predictor step. (default: 30)',
    'num_genetic_cycle': 'number of genetic algorithm cycle (default: 3)',
}

float_kwargs_names = {
}

bool_kwargs_names = {
    'search_internet': 'If True, using "search internet" tools. (default = False)',
    'verbose': 'If True, print intermediate step. (default = True)'
}


class CLICommand:
    """
    Run ChatMOF code

    ex) chatmof run model_name=gpt-3.5-turbo temperature=0.1 verbose=True
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument

        add(
            '--model-name',
            '-m',
            type=str,
            help="OpenAI model name. model_name must be one of [gpt-4, gpt-3.5-turbo, gpt-3.5-turbo-16k]. (default: gpt-4)",
            choices=['gpt-4', 'gpt-3.5-turbo', 'gpt-3.5-turbo-16k'],
            default='gpt-4'
        )
        add(
            '--temperature',
            '-t',
            help='Temperature of LLM models. The lower the temperature, the more accurate the answer, and the higher the temperature, the more variable the answer. (default: 0.1)',
            type=float,
            default=0.1,
        )

        for key, value in str_kwargs_names.items():
            parser.add_argument(f"--{key}", type=str, required=False, help=f"(optional) {value}")

        for key, value in int_kwargs_names.items():
            parser.add_argument(f"--{key}", type=int, required=False, help=f"(optional) {value}")

        for key, value in float_kwargs_names.items():
            parser.add_argument(f"--{key}", type=float, required=False, help=f"(optional) {value}")

        for key, value in bool_kwargs_names.items():
            parser.add_argument(f"--{key}", action='store_true', required=False, help=f"(optional) {value}")

    @staticmethod
    def run(args):
        from chatmof.main import main
        model = args.model_name
        temperature = args.temperature

        kwargs = {}
        for key in chain(str_kwargs_names.keys(), 
                         int_kwargs_names.keys(),
                         float_kwargs_names.keys(),
                         bool_kwargs_names.keys(),
                         ):
            if value := getattr(args, key):
                kwargs[key] = value
        
        main(model=model, temperature=temperature, **kwargs)


import subprocess
from pathlib import Path
import zipfile
from moftransformer.utils.download import _download_file
from chatmof.config import config


class InstallationError(Exception):
    pass


def download_load_model(direc=None):
    if not direc:
        direc = Path(config['model_dir']).parent  # parent directory
    else:
        direc = Path(direc)

    if not direc.is_dir():
        raise ValueError(f"direc must be path for directory, not {direc}")
    if not direc.exists():
        direc.mkdir(parents=True, exist_ok=True)

    # load_model
    link = "https://figshare.com/ndownloader/files/42642739"
    name = 'load_model'
    filename = direc / f"{name}.zip"
    _download_file(link, filename, name)

    if not filename.exists():
        raise InstallationError(f'{name} does not downloaded.')
    
    zip_file = zipfile.ZipFile(str(filename))
    zip_file.extractall(path=direc)


def download_hmof(direc=None):
    if not direc:
        direc = Path(config['hmof_dir']).parent  # parent directory
    else:
        direc = Path(direc)

    if not direc.is_dir():
        raise ValueError(f"direc must be path for directory, not {direc}")
    if not direc.exists():
        direc.mkdir(parents=True, exist_ok=True)

    # load_model
    link = "https://figshare.com/ndownloader/files/42633475"
    name = 'hMOF'
    filename = direc / f"{name}.zip"
    _download_file(link, filename, name)

    if not filename.exists():
        raise InstallationError(f'{name} does not downloaded.')
    
    zip_file = zipfile.ZipFile(str(filename))
    zip_file.extractall(path=direc)


def download_coremof(direc=None):
    if not direc:
        direc = Path(config['data_dir']).parent  # parent directory
    else:
        direc = Path(direc)

    if not direc.is_dir():
        raise ValueError(f"direc must be path for directory, not {direc}")
    if not direc.exists():
        direc.mkdir(parents=True, exist_ok=True)

    # load_model
    link = "https://figshare.com/ndownloader/files/42808069"
    name = 'coremof'
    filename = direc / f"{name}.zip"
    _download_file(link, filename, name)

    if not filename.exists():
        raise InstallationError(f'{name} does not downloaded.')
    
    zip_file = zipfile.ZipFile(str(filename))
    zip_file.extractall(path=direc)


def setup():
    download_coremof()
    download_hmof()
    download_load_model()


import os
from typing import Dict, List
from pathlib import Path
import torch
from moftransformer.datamodules import Dataset as MOFDataset

from chatmof import __root_dir__


class ChatDataset(MOFDataset):
    def __init__(
            self,
            data_list: List[Path],
            nbr_fea_len: int = 64,
    ) -> Dict[str, torch.Tensor]:
        
        super(MOFDataset, self).__init__()  # inheritance TorchDatset

        self.data_list = data_list
        self.draw_false_grid = False
        self.split = ''
        self.nbr_fea_len = nbr_fea_len
        self.tasks = {}

        self.cif_ids = [
            cif.stem for cif in data_list
        ]

        self.data_dir = data_list[0].parent
        
        self.targets = [0] * len(self.cif_ids)



import logging
from typing import Dict, Any, List
from pathlib import Path
from collections import defaultdict
from functools import partial
import yaml

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from moftransformer.modules import Module
from moftransformer.utils.validation import _IS_INTERACTIVE

from chatmof.config import config as default_config
from chatmof.tools.predictor.dataset import ChatDataset


_predictable_properties = [
    path.stem for path in Path(default_config['model_dir']).iterdir() if not path.stem.startswith('__')
]

model_names = ",".join(_predictable_properties + ['else'])


def read_yaml(config: str) -> Dict[str, Any]:
    path = Path(config)
    if not path.exists():
        raise FileNotFoundError(f'Config file does not existed in yaml file, {config}')
    elif path.suffix != '.yaml':
        raise TypeError(f'config file must be *.yaml, not {path.suffix}')
    
    with open(config) as f:
        file = yaml.load(f, Loader=yaml.Loader)

    return file['config']


def update_config(_config: Dict[str, Any], p_model: Path) -> None:
    pl.seed_everything(_config["seed"])
    _config['accelerator'] = default_config['accelerator']
    _config['load_path'] = str(p_model)
    _config['max_epochs'] = 1
    _config['devices'] = 1


def load_datamodule(
        data_list: List[Path], 
        _config: Dict[str, Any]
    ) -> DataLoader:

    dataset = ChatDataset(
        data_list=data_list,
        nbr_fea_len=_config['nbr_fea_len']
    )

    collate_fn = partial(
        ChatDataset.collate,
        img_size=_config['img_size']
    )

    return DataLoader(
        dataset=dataset,
        collate_fn=collate_fn,
        batch_size=_config['per_gpu_batchsize'],
        num_workers=_config['num_workers'],
    )


def load_trainer(_config: Dict[str, Any]) -> pl.Trainer:
    if _IS_INTERACTIVE:
        strategy = None
    elif pl.__version__ >= '2.0.0':
        strategy = "ddp_find_unused_parameters_true"
    else:
        strategy = "ddp"

    trainer = pl.Trainer(
        accelerator=_config["accelerator"],
        devices=_config["devices"],
        num_nodes=_config["num_nodes"],
        precision=_config["precision"],
        strategy=strategy,
        max_epochs=_config["max_epochs"],
        log_every_n_steps=0,
        logger=False,
    )
    return trainer


def predict(
        data_list: List[Path],
        model_dir: Path,
        verbose: bool = False,
) -> Dict[str, List[str]]:
    
    #if not verbose:
    logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)    

    params = model_dir/'hparams.yaml'
    p_model, = model_dir.glob('*.ckpt')
    config = read_yaml(params)
    update_config(config, p_model)

    dataloader = load_datamodule(data_list, config)
    model = Module(config)
    trainer = load_trainer(config)

    rets = trainer.predict(model, dataloader)
    output = defaultdict(list)
    for ret in rets:
        for key, value in ret.items():
            output[key].extend(value)

    return output


def search_file(name: str, direc: Path) -> List[Path]:
    name = name.strip()
    if '*' in name:
        return list(direc.glob(name))
    
    f_name = direc/name
    if f_name.exists():
        return [f_name]
    
    return False


import os
import re
from pathlib import Path
import pandas as pd
from typing import Dict, Any, List, Optional

from langchain.base_language import BaseLanguageModel
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.prompts import PromptTemplate
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.tools.python.tool import PythonAstREPLTool

from chatmof import __root_dir__
from chatmof.config import config
from chatmof.tools.search_csv.base import TableSearcher
from chatmof.tools.predictor.utils import model_names, _predictable_properties
from chatmof.tools.predictor.runner import MOFTransformerRunner
from chatmof.tools.predictor.prompt import (
    PROMPT, FINAL_MARKDOWN_PROPMT
)


class Predictor(Chain):
    """Tools that predict properties using MOFTransformer.
    """
    llm: BaseLanguageModel
    llm_chain: LLMChain
    final_single_chain: LLMChain
    model_dir: str= config['model_dir']
    data_dir: str = config['data_dir']
    tool_names: str = model_names
    input_key: str = 'question'
    output_key: str = 'answer'

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]
    
    def _parse_output(self, text) -> Dict[str, Any]:
        thought = re.search(r"Thought:\s*(.+?)\s*\n", text, re.DOTALL)
        properties = re.findall(r"Property:\s*(.+?)\s*\n", text, re.DOTALL)
        materials = re.findall(r"Material:\s*(.+?)\s*\n", text, re.DOTALL)
        final_thought = re.search(r"Final Thought:\s*(.+?)\s*(\n|$)", text, re.DOTALL)

        if not thought:
            raise ValueError(f'unknown format from LLM: {text}')
        if not properties:
            raise ValueError(f'unknown format from LLM: {text}')
        if not materials:
            raise ValueError(f'unknown format from LLM: {text}')
        if not final_thought:
            raise ValueError(f'unknown format from LLM: {text}')
        if len(properties) != len(materials):
            raise ValueError('number of properties and materials are not the same: {text}')
        for prop in properties:
            if prop not in _predictable_properties:
                raise ValueError(f'Predictor can not predict the property: {prop}')
        
        return {
            'Thought': thought.group(1),
            'Property': properties,
            'Materials': materials,
            'Final Thought': final_thought.group(1),
        }            
    
    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None
    ):
        run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = run_manager.get_child()

        llm_output = self.llm_chain.predict(
            question=inputs[self.input_key],
            callbacks=callbacks
        )

        output = self._parse_output(llm_output)
        
        run_manager.on_text(f"\n[Predictor] Thought: ", verbose=self.verbose)
        run_manager.on_text(output['Thought'], verbose=self.verbose, color='yellow')

        df_ls = []
        info_ls = []
        for prop, mat in zip(output['Property'], output['Materials']):
            run_manager.on_text(f"\n[Predictor] Property: ", verbose=self.verbose)
            run_manager.on_text(prop, verbose=self.verbose, color='yellow')
            run_manager.on_text(f"\n[Predictor] Materials: ", verbose=self.verbose)
            run_manager.on_text(f"{mat}\n", verbose=self.verbose, color='yellow')

            runner = MOFTransformerRunner(
                model_dir=self.model_dir,
                data_dir=self.data_dir,
                verbose=self.verbose,
            )
            cif_id, logits, model_info = runner.run(prop, mat)
            df = pd.DataFrame({'cif_id': cif_id, prop: logits})
            df_ls.append(df)
            info_ls.append(model_info)
            
        df_total = df_ls[0]
        for df in df_ls[1:]:
            df_total.merge(df, on='cif_id', how='outer')

        information = f'Information of models : {info_ls}. If unit or condition are existed, you must include it in the final output.'

        run_manager.on_text(f"[Predictor] Final Thought: ", verbose=self.verbose)
        run_manager.on_text(output['Final Thought'], verbose=self.verbose, color='yellow')
        
        if len(df_total) < config['max_length_in_predictor']:
            final_output = self.final_single_chain.run(
                table = df_total.to_markdown(),
                information=information,
                question = output['Final Thought']
            )
        else:
            searcher = TableSearcher.from_dataframe(
                llm = self.llm,
                dataframe = df_total, 
                verbose = self.verbose
            )
            final_output = searcher.run(question=output['Final Thought'],
                                        information=information,
                                        run_manager=run_manager)
            
        return {self.output_key: final_output}

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: str = PROMPT,
        final_single_prompt: str = FINAL_MARKDOWN_PROPMT,
        **kwargs: Any,
    ) -> Chain:
        template = PromptTemplate(
            template=prompt,
            input_variables=['question'],
            partial_variables={'model_names':model_names}
        )
        fs_template = PromptTemplate(
            template=final_single_prompt,
            input_variables=['table', 'information', 'question']
        )

        llm_chain = LLMChain(llm=llm, prompt=template)
        final_single_chain = LLMChain(llm=llm, prompt=fs_template)

        return cls(
            llm=llm,
            llm_chain=llm_chain, 
            final_single_chain=final_single_chain,
            **kwargs)


if __name__ == '__main__':
    from langchain.chat_models import ChatOpenAI
    from matplotlib import pyplot as plt
    
    llm = ChatOpenAI(temperature=0, model_name='gpt-4')

    prompt = 'Find the structure with the highest hydrogen diffusivity in a dilute system'

    tool = Predictor.from_llm(llm, verbose=True)
    output = tool.run(prompt)
    print (output)



import re
from typing import Dict, Any, List, Optional, Iterable
from langchain.base_language import BaseLanguageModel
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.prompts import BasePromptTemplate, PromptTemplate
from langchain.callbacks.manager import CallbackManagerForChainRun

from chatmof.config import config
from chatmof.tools.genetic_algorithm.prompt import GENETIC_PROMPT


class GeneticAlgorithmChain(Chain):
    llm_chain: LLMChain
    llm: Optional[BaseLanguageModel] = None
    prompt: BasePromptTemplate = GENETIC_PROMPT
    input_key: str = 'question'  #: :meta private:
    output_key: str = 'answer'  #: :meta private:

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key, 'parents']
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    def _process_llm_result(
        self,
        llm_output
    ) -> Dict[str, str]:
        
        children = []
        for child in re.split(r"(\n|,\s?|\s)", llm_output):
            if child := self._evaluate_child(child):
                children.append(child)
        return {self.output_key: children}

    def _evaluate_child(self, child):
        child = child.strip()
        #if gene := re.search(r'[a-z]+\+(N|E)\d+\+(N|E)\d+', child):
        if gene := re.search(r'(N|E)\d+\+(N|E)\d+', child):
            return gene.group(0)
        else:
            return None

    def parse_parents(self, parents:Iterable[Iterable[str]]):
        n_parents = config['num_parents']
        names = []

        for i, parent in enumerate(parents):
            if i> n_parents:
                break
            if len(parent) < 2:
                continue
            elif isinstance(parent, list):
                name, *_ = parent
                name = name.strip()
            elif isinstance(parent, dict):
                name, *_ = parent.values()
            
            names.append(name)

        return ", ".join(names) + "\n"
    
    def _call(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ):
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        _run_manager.on_text(inputs[self.input_key])

        parents = inputs['parents']

        llm_output = self.llm_chain.predict(
            question=inputs[self.input_key],
            parents=self.parse_parents(parents),
            callbacks=_run_manager.get_child(),
        )

        return self._process_llm_result(llm_output,)

    @property
    def _chain_type(self) -> str:
        return "genetic_algorithm_chain"
    
    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: BasePromptTemplate = GENETIC_PROMPT,
        **kwargs: Any,
    ) -> Chain:
        prompt = PromptTemplate(
            input_variables=['question', 'parents'],
            template=prompt,
        )
        llm_chain = LLMChain(llm=llm, prompt=prompt)
        return cls(llm_chain=llm_chain, **kwargs)


import streamlit as st
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from chatmof.agents.agent import ChatMOF
from langchain.callbacks import StdOutCallbackHandler

verbose = True
search_internet = False

llm = ChatOpenAI(temperature=0)
callback_manager = [StdOutCallbackHandler()]

chatmof = ChatMOF.from_llm(
    llm=llm, 
    verbose=verbose, 
)

st.title('Welcome to the ChatMOF 🤖')
input = st.text_input('Question for Metal-organic Framework!')

if input:
    text = chatmof.run(input)
    st.text(text)



from typing import Any, Dict, List, Optional
from pydantic import BaseModel
from langchain.chains.base import Chain
from langchain.base_language import BaseLanguageModel
from langchain.agents import initialize_agent, AgentType
from langchain.callbacks.base import BaseCallbackHandler
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.prompts import PromptTemplate
from chatmof.tools import load_chatmof_tools
from chatmof.agents.prompt import PREFIX, FORMAT_INSTRUCTIONS, SUFFIX


class ChatMOF(Chain):
    llm: BaseLanguageModel
    agent: object
    verbose: bool = False
    input_key: str = "query"
    output_key: str = "output"
    
    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]
    
    def _call(
            self,
            inputs: Dict[str, Any],
    ) -> Dict[str, Any]:
        
        query = inputs[self.input_key]
        output = self.agent.run(query)
        
        return {
            self.output_key: output
        }
    
    @classmethod
    def from_llm(
        cls: BaseModel,
        llm: BaseLanguageModel,
        verbose: bool = False,
        search_internet: bool = True,
    ) -> Chain:
        
        tools = load_chatmof_tools(
            llm=llm, 
            verbose=verbose, 
            search_internet=search_internet
        )
        
        agent_kwargs = {
            'prefix': PREFIX,
            'format_instructions': FORMAT_INSTRUCTIONS,
            'suffix': SUFFIX
        }
        agent = initialize_agent(
            tools,
            llm,
            agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
            verbose=verbose,
            agent_kwargs=agent_kwargs,
            #handle_parsing_errors=True,
        )

        print (type(agent))

        return cls(agent=agent, llm=llm, verbose=verbose)



import os
import re
from pathlib import Path
import pandas as pd
from typing import Dict, Any, List, Optional

import ase.io
import ase.visualize
from langchain.base_language import BaseLanguageModel
from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.prompts import PromptTemplate
from langchain.callbacks.manager import CallbackManagerForChainRun

from chatmof.config import config
from chatmof.utils import search_file
from chatmof.tools.visualizer.prompt import PROMPT


class Visualizer(Chain):
    """Tools that search csv using Pandas agent"""
    llm_chain: LLMChain
    data_dir: Path = Path(config['structure_dir'])
    input_key: str = 'question'
    output_key: str = 'answer'

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]
    
    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]
    
    def _write_log(self, action, text, run_manager):
        run_manager.on_text(f"\n[Visualizer] {action}: ", verbose=self.verbose)
        run_manager.on_text(text, verbose=self.verbose, color='yellow')

    def _parse_output(self, text:str) -> Dict[str, Any]:
        thought = re.search(r"(?<!Final )Thought:\s*(.+?)\s*(Material|Thought|Question|$)", text, re.DOTALL)
        material = re.search(r"Material:\s*(?:```|`)?(.+?)(?:```|`)?\s*(Material|Thought|Question|$)", text, re.DOTALL)

        if not material:
            raise ValueError(f'unknown format for LLM: {text}')
        
        return{
            'Thought': (thought.group(1) if thought else None),
            'Material': (material.group(1) if material else None),
        }

    def _visualize(self, f_st: Path) -> None:
        atoms = ase.io.read(f_st)
        ase.visualize.view(atoms)
    
    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None
    ):
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()
        
        llm_output = self.llm_chain.run(
            question = inputs[self.input_key],
            callbacks=callbacks,
            stop=['Question:'],
        )

        output = self._parse_output(llm_output)
        self._write_log('Thought', output['Thought'], run_manager)
        self._write_log('Material', output['Material'], run_manager)

        materials = output['Material'].split(',')
        for material in materials:
            material = material.strip()
            ls_st = search_file(f'**/**/{material}.cif', self.data_dir)
            if not ls_st:
                return {self.output_key: f'There are no structures with {material}.'}
            
            if '*' in material:
                for f_st in ls_st:
                    self._visualize(f_st)
            else:
                f_st = ls_st[0]
                self._visualize(f_st)

        return {self.output_key: f'The visualizer has successfully visualized the structure {materials}.'}
    
    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: str = PROMPT,
        **kwargs
    ) -> Chain:
        template = PromptTemplate(
            template=prompt,
            input_variables=['question'],
        )
        llm_chain = LLMChain(llm=llm, prompt=template)
        return cls(llm_chain=llm_chain, **kwargs)


ASE_PROMPT = """You are working with a ase library in Python. The name of the object `ase.atoms` is `atoms`.
You should make a valid python command as input. You must use print the output using the `print` function at the end.

Use the following format:

Question: the input question you must answer
Thought: you should always think about what to do
Input: the valid python code only using the ase library
Observation: the result of python code
... (this Thought/Input/Observation can repeat N times)
Final Thought: you should think about how to answer the question based on your observation
Final Answer: the final answer to the original input question. If you can't answer the question, say `nothing`


Begin!

Question: What is the head of df? If you extracted successfully, derive 'success' as the final answer
Thought: To get the head of a DataFrame, we can use the pandas function head(), which will return the first N rows. By default, it returns the first 5 rows.
Input: 
``` 
import pandas as pd
import json
print(df.head().to_markdown())
```
Observation: {df_head}
Final Thought: The head() function in pandas provides the first 5 rows of the DataFrame. 
Final Answer: success


Question: {question}
{agent_scratchpad}
"""


from typing import Any, Dict
from pathlib import Path
from pydantic import BaseModel, root_validator
from langchain.tools import Tool
from langchain.base_language import BaseLanguageModel
from langchain.tools.base import BaseTool
from chatmof import __root_dir__
from chatmof.tools.genetic_algorithm.base import Generator


def _get_generator(
        llm: BaseLanguageModel,
        verbose: bool = False,
        **kwargs: Any
) -> BaseTool:
    return Tool(
        name="generator",
        description=(
            "The tool to use when you need to create materials with specific properties. "
            "`generator` must only be used when the original question wants to generate. "
            "input must be provided in the form of a full sentence. "
        ),
        func=Generator.from_llm(llm=llm, verbose=verbose).run,
    )


import warnings
from typing import Dict, Callable, List
from pydantic import ValidationError
from langchain.base_language import BaseLanguageModel
from langchain.tools.base import BaseTool
from langchain.agents import load_tools, get_all_tool_names

from chatmof.tools.predictor import _get_predict_properties
from chatmof.tools.search_csv import _get_search_csv
from chatmof.tools.genetic_algorithm import _get_generator
from chatmof.tools.visualizer import _get_visualizer
from chatmof.tools.python_repl import _get_python_repl


_MOF_TOOLS: Dict[str, Callable[[BaseLanguageModel], BaseTool]] = {
    "search_csv" : _get_search_csv,
    "predictor": _get_predict_properties,
    'generator': _get_generator,
    "visualizer": _get_visualizer,
    "python_repl": _get_python_repl,
}

_load_tool_names: List[str] = [
    #'python_repl',
    #'requests',
    #'requests_get',
    #'requests_post',
    #'requests_patch',
    #'requests_put',
    #'requests_delete',
    #'terminal',
    #'human',
    'llm-math',
    #'pal-math'
]

_load_internet_tool_names: list[str] = [
    'google-search',
    'google-search-results-json',
    'wikipedia',
]


def load_chatmof_tools(
        llm: BaseLanguageModel, 
        verbose: bool=False,
        search_internet: bool=True,    
    ) -> List[BaseTool]:
    tools = load_tools(_load_tool_names, llm=llm)
    custom_tools = [model(llm=llm, verbose=verbose) for model in _MOF_TOOLS.values()]

    if search_internet:
        try:
            internet_tools = load_tools(_load_internet_tool_names, llm=llm)
        except Exception as e:
            warnings.warn(f'The internet tool does not work: {e}')
            internet_tools = []
        return custom_tools + tools + internet_tools
    else:
        return custom_tools + tools


import re
from setuptools import setup, find_packages


with open("requirements.txt", "r") as f:
    install_requires = f.readlines()

with open("README.md", "r") as f:
    long_description = f.read()

extras_require = {"docs": ["sphinx", "livereload", "myst-parser"]}

with open("chatmof/__init__.py") as f:
    version = re.search(r"__version__ = [\'\"](?P<version>.+)[\'\"]", f.read()).group("version")


setup(
    name="chatmof",
    version=version,
    description="chatmof",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Yeonghun Kang",
    author_email="dudgns1675@kaist.ac.kr",
    packages=find_packages(),
    package_data={
        'chatmof': [
            'database/tables/*.xlsx'
        ]
    },
    install_requires=install_requires,
    extras_require=extras_require,
    scripts=[],
    #url="https://hspark1212.github.io/MOFTransformer/",
    download_url="https://github.com/Yeonghun1675/ChatMOF",
    entry_points={"console_scripts": ["chatmof=chatmof.cli.main:main"]},
    python_requires=">=3.9",
)
